"""Tests that the approaches for ensemble model inference are correct.
"""


import pytest
import numpy as np
from ml_uncertainty.model_inference import EnsembleModelInference
from sklearn.ensemble import RandomForestRegressor

np.random.seed(1)


@pytest.fixture()
def set_up_random_forest():
    """Sets up RF for benchmarking."""

    X = (np.random.random(size=10000) * 10 + 5).reshape((-1, 1))

    int_for_split = int(0.8 * X.shape[0])
    X_train = X[:int_for_split, :]
    X_test = X[int_for_split:, :]

    def model(X, coef_):
        """ """
        return np.dot(X, coef_)

    # Training y
    true_params = np.array([1.0])
    noise = np.random.normal(loc=0, scale=1, size=X.shape[0])
    y = model(X, true_params) + noise

    y_train = y[:int_for_split]
    y_test = y[int_for_split:]

    # Be sure to have a large number of trees (n_estimators)
    # and max_samples < 1, if we want to analyze the train OOB features.
    regr = RandomForestRegressor(
        max_depth=10, random_state=0, max_samples=0.7, n_estimators=500
    )

    regr.fit(X_train, y_train)

    y_pred_train = regr.predict(X_train)
    y_pred_test = regr.predict(X_test)

    return (
        X,
        y,
        X_train,
        y_train,
        X_test,
        y_test,
        true_params,
        noise,
        y,
        regr,
        y_pred_train,
        y_pred_test,
    )


def compute_coverage(df_int, y):
    """Gets coverage from the dataframe."""

    lower_bound = df_int["lower_bound"].values
    upper_bound = df_int["upper_bound"].values

    coverage = ((lower_bound < y) & (upper_bound > y)).sum() / y.shape[0]

    return coverage


def test_prediction_interval_marginal(set_up_random_forest):
    """Computes prediction interval on the model and checks
    that it has the right coverage.
    Uses the 'marginal' method to obtain the prediction interval.
    """

    (
        X,
        y,
        X_train,
        y_train,
        X_test,
        y_test,
        true_params,
        noise,
        y,
        regr,
        y_pred_train,
        y_pred_test,
    ) = set_up_random_forest

    inf = EnsembleModelInference()

    inf.set_up_model_inference(X_train, y_train, regr, variance_type_to_use="marginal")

    # Gets prediction intervals NOT using SD.
    df_int_list_1 = inf.get_intervals(
        X_train,
        is_train_data=True,
        type_="prediction",
        estimate_from_SD=False,
        confidence_level=90.0,
        return_full_distribution=False,
    )

    df_int_1 = df_int_list_1[0]
    coverage_1 = compute_coverage(df_int_1, y_train)

    # Test when prediction  intervals are from SD.
    df_int_list_2 = inf.get_intervals(
        X_train,
        is_train_data=True,
        type_="prediction",
        estimate_from_SD=True,
        confidence_level=90.0,
        return_full_distribution=False,
    )
    df_int_2 = df_int_list_2[0]
    coverage_2 = compute_coverage(df_int_2, y_train)

    # Tests coverage
    assert (
        np.abs(coverage_1 - 0.90) < 0.02
    ), "Coverage for marginal with non-SD\
            deviates from confidence_level."

    assert (
        np.abs(coverage_2 - 0.90) < 0.02
    ), "Coverage for marginal with SD\
            deviates from confidence_level."


def test_prediction_interval_individual(set_up_random_forest):
    """Computes prediction interval on the model and checks
    that it has the right coverage.
    Uses the individual trees to obtain an estimate of the prediction
    intervals.
    """

    (
        X,
        y,
        X_train,
        y_train,
        X_test,
        y_test,
        true_params,
        noise,
        y,
        regr,
        y_pred_train,
        y_pred_test,
    ) = set_up_random_forest

    inf = EnsembleModelInference()

    inf.set_up_model_inference(
        X_train, y_train, regr, variance_type_to_use="individual"
    )

    # Gets prediction intervals NOT using SD.
    df_int_list_1 = inf.get_intervals(
        X_train,
        is_train_data=True,
        type_="prediction",
        estimate_from_SD=False,
        confidence_level=90.0,
        return_full_distribution=False,
    )
    df_int_1 = df_int_list_1[0]
    coverage_1 = compute_coverage(df_int_1, y_train)

    # Test when prediction  intervals are from SD.
    df_int_list_2 = inf.get_intervals(
        X_train,
        is_train_data=True,
        type_="prediction",
        estimate_from_SD=True,
        confidence_level=90.0,
        return_full_distribution=False,
    )
    df_int_2 = df_int_list_2[0]
    coverage_2 = compute_coverage(df_int_2, y_train)

    # Expected coverages

    # Tests coverage
    assert (
        np.abs(coverage_1 - 0.95) < 0.02
    ), "Coverage for individual with non-SD\
            deviates from expected value."

    assert (
        np.abs(coverage_2 - 0.95) < 0.02
    ), "Coverage for individual with SD\
            deviates from expected value."


def test_confidence_intervals(set_up_random_forest):
    """Computes confidence intervals."""

    (
        X,
        y,
        X_train,
        y_train,
        X_test,
        y_test,
        true_params,
        noise,
        y,
        regr,
        y_pred_train,
        y_pred_test,
    ) = set_up_random_forest

    inf = EnsembleModelInference()

    inf.set_up_model_inference(
        X_train, y_train, regr, variance_type_to_use="individual"
    )

    # Gets prediction intervals NOT using SD.
    df_int_list_1 = inf.get_intervals(
        X_train,
        is_train_data=True,
        type_="confidence",
        estimate_from_SD=False,
        confidence_level=90.0,
        return_full_distribution=False,
    )
    df_int_1 = df_int_list_1[0]
    coverage_1 = compute_coverage(df_int_1, y_train)

    # Test when prediction  intervals are from SD.
    df_int_list_2 = inf.get_intervals(
        X_train,
        is_train_data=True,
        type_="confidence",
        estimate_from_SD=True,
        confidence_level=90.0,
        return_full_distribution=False,
    )
    df_int_2 = df_int_list_2[0]
    coverage_2 = compute_coverage(df_int_2, y_train)

    # Expected coverages

    # Tests coverage against expected coverage
    assert (
        np.abs(coverage_1 - 0.47) < 0.02
    ), "Coverage for individual with non-SD\
            deviates from expected value."

    assert (
        np.abs(coverage_2 - 0.50) < 0.02
    ), "Coverage for individual with SD\
            deviates from expected value."



"""We benchmark SM examples from website.
https://www.statsmodels.org/dev/examples/notebooks/generated/ols.html

NOTE: This example is not added to tests. as this example and sm_linear_models.py
tackle almost the same problem.
"""

#%%
# Statsmodels example

import numpy as np
import statsmodels.api as sm
from ml_uncertainty.non_linear_regression import NonLinearRegression
from ml_uncertainty.model_inference import ParametricModelInference

np.random.seed(9876789)

# Statsmodels example
nsample = 100
x = np.linspace(0, 10, 100)
X = np.column_stack((x, x ** 2))
beta = np.array([1, 0.1, 10])
e = np.random.normal(size=nsample)

X = sm.add_constant(X)
y = np.dot(X, beta) + e

model = sm.OLS(y, X)
results = model.fit()
# print(results.summary())
sm_params = results.params
sm_param_se = np.sqrt(np.diag(results.cov_params()))  # parameter error
sm_pred_df = results.get_prediction().summary_frame(alpha=0.05)

#%%
# NLR example


# Create the model.
def model(X_arr, coefs_):
    """ """
    return coefs_[0] + coefs_[1] * X_arr[:, 0] + coefs_[2] * X_arr[:, 0] ** 2


X_arr = x.reshape((-1, 1))
y_model = model(X_arr, beta)

# Fit using NLR
nlr = NonLinearRegression(model=model, p0_length=beta.shape[0], fit_intercept=True)

nlr.fit(X_arr, y)

inf = ParametricModelInference()

inf.set_up_model_inference(X_arr, y, nlr)

# Get prediction intervals on the features.
df_feature_imp = inf.get_parameter_errors()

# Get prediction intervals.
# Getting prediction intervals for X_expt values.
df_int = inf.get_intervals(X_arr, confidence_level=95.0, distribution="t")

# Compare sm_best_fit_params with nlr.coef_
assert np.linalg.norm(sm_params - nlr.coef_) < 1e-3, "Best fit parameters not equal"

# Compare pred_sm_df with df_feature_imp
assert (
    np.linalg.norm(sm_param_se - df_feature_imp["std"].values) < 1e-3
), "Stanard error of parameter values are different"

# Compare pred_sm_df with df_int
assert (
    np.linalg.norm(
        (
            sm_pred_df[["obs_ci_lower", "obs_ci_upper"]]
            - df_int[["lower_bound", "upper_bound"]].values
        )
        / df_int.shape[0]
    )
    < 1e-3
), "Prediction intervals predicted are different"


# %%



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.md') as history_file:
    history = history_file.read()

with open("requirements.txt") as requirements_file:
    requirement_list = requirements_file.readlines()

requirements = requirement_list
#requirements = [ ]

test_requirements = ['pytest==6.2.4', "click==8.0.3", 
                     "black==21.7b0", "flake8==3.7.8"]

setup(
    author="Archit Nikhil Datar",
    author_email='architdatar@gmail.com',
    python_requires='>=3.9,<=3.12',
    classifiers=[
        "Development Status :: 3 - Alpha",
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    description="Uncertainty quantification and model inference for machine learning models",
    entry_points={
        'console_scripts': [
            'ml_uncertainty=ml_uncertainty.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='ml_uncertainty',
    name='ml_uncertainty',
    packages=find_packages(include=['ml_uncertainty', 'ml_uncertainty.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/architdatar/ml_uncertainty',
    version='0.1.1',
    zip_safe=False,
)



"""Contains statisical utils utilized by several classes"""

from scipy.stats import norm, t as t_dist
import numpy as np


def get_significance_levels(confidence_level, side):
    """Computes the significance level for the given side

    Effectively, returns the percentiles of the distributions which we wish to
    evaluate.

    In other words, it is the stat at which the percent of distribution
    less than the desired value is the desired probability.

    Parameters
    ----------
    confidence_level: float
        Confidence level desired.
    side: str, {"two-sided", "lower", "upper"}
        Side of the interval we want

    Returns
    --------
    significance_level: float
        Stats between which our desired interval should lie.

    Examples:
    --------
    1. For a 2-sided 90% confidence interval, we would wish to obtain the stats
        x1* and x2* such that P(x1* < x < x2*) = 90%..
        Considering a symmetric distribution: we would want
        P(x < x1*) = 5% and P(x > x2*) = 5% => P(x < x2*) = 95%
    2. For a lower 90%, we would wish to obtain the stat
        x* such that P(x >= x*) = 90%, alternatively, P(x < x*) = 10%
    3. For upper 90% CI, we would wish to obtain the stat
        x* such that P(x < x*) = 90%
    """

    if side == "two-sided":
        significance_levels = [
            (1 - confidence_level / 100) / 2,
            1 - (1 - confidence_level / 100) / 2,
        ]
    elif side == "upper":
        significance_levels = [1 - confidence_level / 100, None]
    elif side == "lower":
        significance_levels = [None, confidence_level / 100]
    else:
        raise ValueError(
            "Side should be one of \
            two-sided, lower, and upper. Please provided one of these."
        )

    return significance_levels


def get_z_values(significance_levels):
    """Gets z-stats from normal distribution for the desired
    significance levels.

    Parameters
    ----------
    significance_levels: list
        Significance levels

    Returns
    -------
    list
        Z-values
    """
    # Gets z values for the required significance_levels
    z_values = []
    for level in significance_levels:
        if level is None:
            z_values.append(None)
        else:
            stat = norm.ppf(level)
            z_values.append(stat)

    return z_values


def get_t_values(significance_levels, dfe):
    """Computes t values

    Parameters
    ----------
    significance_levels: list
        Significance levels
    dfe: int
        Degrees of freedom

    Returns
    -------
    list
        t-values
    """

    if not dfe:
        # dfe is not supplied.
        raise ValueError(
            "dfe is not supplied\
                due to which t-stats cannot be computed.\
                If you wish to make a large sample approximation (LSA)\
                that your sample standard deviation is based on enough samples to\
                be close to the population standard deviation, you may consider using\
                the normal distribution instead."
        )
    else:
        # For each level: get the desired t-stat.
        t_values = []
        for level in significance_levels:
            if level is None:
                t_values.append(None)
            else:
                stat = t_dist.ppf(q=level, df=dfe)
                t_values.append(stat)

        return t_values


def compute_intervals(
    mean,
    se,
    side="two-sided",
    confidence_level=95.0,
    distribution="normal",
    dfe=None,
):
    """Computes the desired intervals given the mean and standard error (SE) values.

    Parameters:
    ------------
    mean: array
        Means of the distributions
    se: array
        Standard errors
    side: str, optional, {"two-sided", "lower", "upper"}, default: "two-sided"
        Defines the type of interval to be computed.
    confidence_level: float, optional, default: 90
        Percentage of the distribution to be enveloped by the interval.
        Example: A value of 90 means that you wish to compute a 90% interval.
    distribution: str, optional, {"normal", "t"}, default: "normal"
        The type of distribution that the desired interval is from.
        For now, only normal and t distributions are supported.
        If you wish to compute an interval on a different distribution, please compute
        the appropriate interval externally.
    dfe: {None, float}, optional, default: None
        Only used if distribution == "t" to supply the error degrees of freedom.

    Returns
    ----------
    list of 2 arrays each of shape (n_samples,) containing the lower and upper bounds
      of the desired interval.
    """

    # Gets significance level.
    significance_levels = get_significance_levels(confidence_level, side)

    # Currently, we only provide support for
    # normal and t-distributions.
    if distribution == "normal":
        stats = get_z_values(significance_levels)
    elif distribution == "t":
        stats = get_t_values(significance_levels, dfe)
    else:
        raise ValueError(
            "Currently intervals can only be obtained for normal\
            and t-distributions. In case your data come from a difference \
            distribution, please compute the appropriate intervals externally \
            using y_hat and se."
        )

    # Finally, return desired interval.
    if side == "two-sided":
        return [mean + stats[0] * se, mean + stats[1] * se]
    if side == "upper":
        return [mean + stats[0] * se, np.repeat(np.inf, mean.shape[0])]
    if side == "lower":
        return [-np.repeat(np.inf, mean.shape[0]), mean + stats[1] * se]



"""Benchmarks the non-linear regression against statsmodels results.

This benchmark example creates files that is used in the `test_benchmark_sm_ols.py`
test.
"""

#%%
# Import libraries.
import statsmodels.api as sm
import autograd.numpy as np
from ml_uncertainty.non_linear_regression import NonLinearRegression
from ml_uncertainty.model_inference import ParametricModelInference

np.random.seed(1)


def linear_model(X, beta):
    """ """
    return X @ beta


# Case 1: 1D X. Generate X and y variables
X_expt = np.linspace(0, 10, 1000).reshape((-1, 1))
true_params = np.array([1.0])

# Case 2: 2D X
# X_expt, _ = make_regression(n_samples=500,
#                             n_features=2,
#                             n_informative=2,
#                             random_state=1
#                                  )
# true_params = np.array([1., 1.])


X_expt = X_expt + 500


y_expt = linear_model(X_expt, true_params) + np.random.normal(
    loc=0, scale=1, size=X_expt.shape[0]
)

model = sm.OLS(y_expt, X_expt)

results = model.fit()
sm_best_fit_params = results.params
param_std_error_sm = np.sqrt(np.diag(results.cov_params()))  # parameter error
pred_sm_df = results.get_prediction(X_expt).summary_frame(alpha=0.05)[
    ["obs_ci_lower", "obs_ci_upper"]
]


# Fit the same data using NLR
nlr = NonLinearRegression(model=linear_model, p0_length=true_params.shape[0])

nlr.fit(X_expt, y_expt)

# Predicted with fitted parameters.
y_pred = nlr.predict(X_expt)

inf = ParametricModelInference()

inf.set_up_model_inference(X_expt, y_expt, nlr)

# Get prediction intervals on the features.
df_feature_imp = inf.get_parameter_errors()

# Getting prediction intervals for X_expt values.
df_int = inf.get_intervals(X_expt, confidence_level=95.0, distribution="t")

# Compare sm_best_fit_params with nlr.coef_
assert (
    np.linalg.norm(sm_best_fit_params - nlr.coef_) < 1e-3
), "Best fit parameters not equal"

# Compare pred_sm_df with df_feature_imp
assert (
    np.linalg.norm(param_std_error_sm - df_feature_imp["std"].values) < 1e-3
), "Stanard error of parameter values are different"

# Compare pred_sm_df with df_int
assert (
    np.linalg.norm(
        (pred_sm_df.values - df_int[["lower_bound", "upper_bound"]].values)
        / df_int.shape[0]
    )
    < 1e-3
), "Prediction intervals predicted are different"


# Write out these arrays to a file.
# 1D case
# np.savetxt("sm_outputs/1D_best_fit_params.csv", sm_best_fit_params, fmt="%.10f",
#  delimiter=",")
# np.savetxt("sm_outputs/1D_param_errors.csv", param_std_error_sm, fmt="%.10f",
# delimiter=",")
# np.savetxt("sm_outputs/1D_pred_bounds.csv", pred_sm_df, fmt="%.10f", delimiter=",")

# 2D case
# np.savetxt("sm_outputs/2D_best_fit_params.csv", sm_best_fit_params, fmt="%.10f",
# delimiter=",")
# np.savetxt("sm_outputs/2D_param_errors.csv", param_std_error_sm, fmt="%.10f",
# delimiter=",")
# np.savetxt("sm_outputs/2D_pred_bounds.csv", pred_sm_df, fmt="%.10f", delimiter=",")
